// Copyright 2023 The Chromium Authors
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "third_party/blink/renderer/core/paint/line_oriented_rect.h"
#include "third_party/blink/renderer/core/layout/geometry/physical_rect.h"

#include "testing/gtest/include/gtest/gtest.h"
#include "third_party/blink/renderer/core/testing/core_unit_test_helper.h"

namespace blink {

// TODO:
// What is this functions conceptual purpose?
//   It is only being used within the context of MaybeNotAtBoxOrigin
//   return EnclosingRect(rotated rect)
TEST(LineOrientedRectTest, EnclosingRect) {
  gfx::RectF r(0, 10, 20, 30);
  LineOrientedRect lor = LineOrientedRect::EnclosingRect(r);
  EXPECT_EQ(lor.offset.x, 0) << "";
  EXPECT_EQ(lor.offset.y, 10) << "";
  EXPECT_EQ(lor.size.inline_size, 20) << "";
  EXPECT_EQ(lor.size.block_size, 30) << "";
}

TEST(LineOrientedRectTest, AtBoxOrigin) {
  PhysicalRect r(0, 0, 20, 30);
  LineOrientedRect lor = LineOrientedRect::AtBoxOrigin(r, true);
  EXPECT_EQ(lor.offset.x, 0) << "";
  EXPECT_EQ(lor.offset.y, 0) << "";
  EXPECT_EQ(lor.size.inline_size, 20) << "";
  EXPECT_EQ(lor.size.block_size, 30) << "";

  LineOrientedRect lor_veritcal = LineOrientedRect::AtBoxOrigin(r, false);
  EXPECT_EQ(lor_veritcal.offset.x, 0) << "";
  EXPECT_EQ(lor_veritcal.offset.y, 0) << "";
  EXPECT_EQ(lor_veritcal.size.inline_size, 30) << "";
  EXPECT_EQ(lor_veritcal.size.block_size, 20) << "";
}

// TODO: see "Fails"
// This test fails, because LineRotation does not to what I expect it to do when
// the WritingMode is kHorizontalTb -- I'd expect it to return the idenfity
TEST(LineOrientedRectTest, MaybeNotAtBoxOrigin_kHorizontalTB) {
  PhysicalRect r(0, 0, 20, 30);

  const WritingMode writing_mode = WritingMode::kHorizontalTb;
  const bool is_horizontal = IsHorizontalWritingMode(writing_mode);

  const LineOrientedRect rotated_box =
      LineOrientedRect::AtBoxOrigin(r, is_horizontal);
  absl::optional<AffineTransform> rotation = LineRotation(rotated_box, writing_mode);

  // Fails: Does not return the identify, returns the clockwise rotation
  EXPECT_EQ(rotation, AffineTransform());

  // First half of original box r
  PhysicalRect highlight(0, 0, 20, 15);
  LineOrientedRect rotated =
    LineOrientedRect::MaybeNotAtBoxOrigin(highlight, rotation);
  EXPECT_EQ(rotated.offset.x, 0) << "";
  EXPECT_EQ(rotated.offset.y, 0) << "";
  EXPECT_EQ(rotated.size.inline_size, 15) << "";
  EXPECT_EQ(rotated.size.block_size, 20) << "";

  // Second half of original box r
  PhysicalRect highlight2(0, 15, 20, 15);
  LineOrientedRect rotated2 =
    LineOrientedRect::MaybeNotAtBoxOrigin(highlight2, rotation);
  EXPECT_EQ(rotated2.offset.x, 0) << "";
  EXPECT_EQ(rotated2.offset.y, 15) << "";
  EXPECT_EQ(rotated2.size.inline_size, 20) << "";
  EXPECT_EQ(rotated2.size.block_size, 15) << "";

}

// TODO: See "Fails"
// To get the expectation values, I did the following:
//   1. Drew a box on graph paper (A: (0, 0), B: (0, 30), C: (20, 30), D: (20, 0))
//   2. Rotated by -90 around 0, 0
//   3. Translated until point B is at (0, 0), which is a translation in the y by 20
//   4. Used the same translation steps for the highlight and highlight2
TEST(LineOrientedRectTest, MaybeNotAtBoxOrigin_kSidewaysLr) {
  PhysicalRect r(0, 0, 20, 30);

  const WritingMode writing_mode = WritingMode::kSidewaysLr;
  const bool is_horizontal = IsHorizontalWritingMode(writing_mode);
  const LineOrientedRect rotated_box =
      LineOrientedRect::AtBoxOrigin(r, is_horizontal);
  absl::optional<AffineTransform> rotation = LineRotation(rotated_box, writing_mode);

  // AffineTransform ("translation(0,20), scale(1,1), angle(-90deg), skewxy(0)")
  EXPECT_EQ(rotation, AffineTransform(0, -1, 1, 0, 0, 20));

  // Top half of original box r
  PhysicalRect highlight(0, 0, 20, 15);
  LineOrientedRect rotated =
    LineOrientedRect::MaybeNotAtBoxOrigin(highlight, rotation);
  EXPECT_EQ(rotated.offset.x, 0) << ""; // Fails, calculated: 5 !!!!!!!!!!!!!
  EXPECT_EQ(rotated.offset.y, 0) << "";
  EXPECT_EQ(rotated.size.inline_size, 15) << "";
  EXPECT_EQ(rotated.size.block_size, 20) << "";

  // Bottom half of original box r
  PhysicalRect highlight2(0, 15, 20, 15);
  LineOrientedRect rotated2 =
    LineOrientedRect::MaybeNotAtBoxOrigin(highlight2, rotation);
  EXPECT_EQ(rotated2.offset.x, 15) << ""; // Fails, calculated: -10 !!!!!!!!!!!!!
  EXPECT_EQ(rotated2.offset.y, 0) << "";
  EXPECT_EQ(rotated2.size.inline_size, 15) << "";
  EXPECT_EQ(rotated2.size.block_size, 20) << "";
}

// TODO: See "fails"
// To get the expectation values, I did the following:
//   1. Drew a box on graph paper (A: (0, 0), B: (0, 30), C: (20, 30), D: (20, 0))
//   2. Rotated by 90 around 0, 0
//   3. Translated until point B is at (0, 0), which is a translation in the x by 30
//      SO, the rotation I expect is: ("translation(30,0), scale(1,1), angle(90deg), skewxy(0)")
//      BUT, I got:                   ("translation(20,0), scale(1,1), angle(90deg), skewxy(0)")
//   4. Used the same translation steps for the highlight and highlight2 boxes
TEST(LineOrientedRectTest, MaybeNotAtBoxOrigin_kVerticalRl) {
  PhysicalRect r(0, 0, 20, 30);

  const WritingMode writing_mode = WritingMode::kVerticalRl;
  const bool is_horizontal = IsHorizontalWritingMode(writing_mode);
  const LineOrientedRect rotated_box =
      LineOrientedRect::AtBoxOrigin(r, is_horizontal);
  absl::optional<AffineTransform> rotation = LineRotation(rotated_box, writing_mode);

  // AffineTransform ("translation(30,0), scale(1,1), angle(90deg), skewxy(0)")
  EXPECT_EQ(rotation, AffineTransform(0, 1, -1, 0, 30, 0));
  // Fails, calculated: AffineTransform(0, 1, -1, 0, 20, 0)); !!!!!!!!!!!!!!

  // Top half of original box r
  PhysicalRect highlight(0, 0, 20, 15);
  LineOrientedRect rotated =
    LineOrientedRect::MaybeNotAtBoxOrigin(highlight, rotation);
  EXPECT_EQ(rotated.offset.x, 15) << ""; //  Fails, calculated: 0 !!!!!!!!!!!!!
  EXPECT_EQ(rotated.offset.y, 0) << "";
  EXPECT_EQ(rotated.size.inline_size, 15) << "";
  EXPECT_EQ(rotated.size.block_size, 20) << "";

  // Bottom half of original box r
  PhysicalRect highlight2(0, 15, 20, 15);
  LineOrientedRect rotated2 =
    LineOrientedRect::MaybeNotAtBoxOrigin(highlight2, rotation);
  EXPECT_EQ(rotated2.offset.x, 0) << ""; //  Fails, calculated: 15 !!!!!!!!!!!!!
  EXPECT_EQ(rotated2.offset.y, 0) << "";
  EXPECT_EQ(rotated2.size.inline_size, 15) << "";
  EXPECT_EQ(rotated2.size.block_size, 20) << "";
}

}  // namespace blink
